""" secure a users file """
from invoke import task
from dotenv import load_dotenv
from cryptography.fernet import Fernet
from duckdown.utils.encrypt import encrypt, decrypt
from duckdown.utils import json_utils


@task
def secure(_, path):
    """ encrypt the passwords in a users.json file """
    load_dotenv(verbose=True)

    with open(path) as file:
        users = json_utils.load(file)

    new_values = {}
    for user in users:
        new_values[user] = encrypt(user)

    print(json_utils.dumps(new_values))


@task
def unsecure(_, data):
    """ return to normal """
    load_dotenv(verbose=True)
    print(decrypt(data))


@task
def gen_key(_):
    """ generates a secret key  for use as DKDN_KEY envar"""
    return Fernet.generate_key()



# pylint: disable=unused-variable, too-many-locals
""" application entry point """
import logging
import os
import tornado.ioloop
import tornado.web
import tornado.log
import boto3
from .utils import json_utils
from .utils.folder import Folder
from .utils.s3_folders import S3Folder
from .handlers.access_control import DictAuthenticator
from . import handlers

LOGGER = logging.getLogger(__name__)


class App(tornado.web.Application):
    """ an app served up out of a directory """

    def __init__(self, cfg, routes=None, **settings):
        """ set up """
        self.cfg = cfg

        s3client = None
        s3region = None

        if cfg.bucket_name:
            LOGGER.info("duckdown s3: %s", cfg.bucket_name)
            s3client, s3region = self.make_duck_s3client(**cfg.credentials)

            self.folder = S3Folder(
                cfg.bucket_name, client=s3client, region=s3region
            )
            cfg.convert_image_paths = False
            cfg.image_path = False
            cfg.image_bucket = (
                cfg.image_bucket if cfg.image_bucket else cfg.bucket_name
            )
            cfg.img_path = f"{self.folder.s3bucket_url}/{cfg.IMAGES_PATH}"
        elif cfg.app_path:
            LOGGER.info("duckdown local: %s", cfg.app_path)
            self.folder = Folder(directory=cfg.app_path)
        else:
            raise Exception("app_path or bucket required")

        if cfg.image_bucket:
            cfg.convert_image_paths = False
            cfg.image_path = False
            if s3client is None:
                s3client, s3region = self.make_duck_s3client(
                    **cfg.image_credentials
                )
            image_folder = S3Folder(
                cfg.image_bucket, client=s3client, region=s3region
            )
            cfg.img_path = f"{image_folder.s3bucket_url}/{cfg.IMAGES_PATH}"
            self.folder.set_image_bucket(image_folder)

        settings = cfg.tornado_settings(settings or {})
        settings.setdefault("static_handler_class", handlers.StaticFiles)
        routes = self.setup_duck_routes(cfg, routes)
        super().__init__(
            routes,
            **settings,
        )

    def load_users(self, request):
        """ return the user and password """
        site = self.get_site(None, None, request)
        users_path = self.settings["users_path"]
        LOGGER.info("loading users: %s", users_path)
        return DictAuthenticator(
            json_utils.loads(site.get_file(users_path)[-1])
        )

    def get_site(
        self, user=None, path=None, request=None
    ):  # pylint: disable=unused-argument
        """ return current site """
        return self.folder

    def setup_duck_routes(self, cfg, routes):
        """ do the thing """
        routes = routes or []

        login_handler = cfg.login_handler or (handlers.LoginHandler,)
        routes.extend(
            [
                (r"/login", *login_handler),
                (r"/logout", handlers.LogoutHandler),
                (
                    r"/edit/browse/(.*)",
                    handlers.S3Browser,
                    {
                        "bucket_name": cfg.image_bucket,
                        "folder": cfg.image_path,
                    },
                ),
                (
                    r"/edit/assets/(.*)",
                    tornado.web.StaticFileHandler,
                    {"path": cfg.duck_assets},
                ),
                (r"/edit/mark/", handlers.MarkHandler),
                (
                    r"/edit/pages/(.*)",
                    handlers.DirHandler,
                    {"directory": cfg.PAGE_PATH},
                ),
                (
                    r"/edit",
                    handlers.EditorHandler,
                    {
                        "page": cfg.vue_page,
                        "manifest": self.load_vue_manifest(
                            cfg.vue_manifest, cfg.debug
                        ),
                    },
                ),
            ]
        )

        if cfg.debug is True:
            LOGGER.info("vue dev handler: %s", cfg.vue_src)
            routes.insert(
                0,
                (
                    r"/src/(.*)",
                    tornado.web.StaticFileHandler,
                    {"path": cfg.vue_src},
                ),
            )
        else:
            LOGGER.debug("vue manifest handler")
            routes.insert(
                0,
                (
                    r"/_assets/(.*)",
                    tornado.web.StaticFileHandler,
                    {"path": cfg.vue_assets},
                ),
            )

        routes.append(
            (
                r"/(.*)",
                handlers.SiteHandler,
                {"pages": cfg.PAGE_PATH},
            )
        )
        return routes

    @classmethod
    def load_vue_manifest(cls, path, debug):
        """ loading manifest for vue dev environment """
        manifest = None
        if debug is False:
            LOGGER.info("loading vue manifest: %s", path)
            with open(path) as file:
                manifest = json_utils.load(file)
        return manifest

    @classmethod
    def make_duck_s3client(cls, **credentials):
        """ determine region or get from env """
        s3region = (
            credentials["region_name"]
            if "region_name" in credentials
            else os.environ["AWS_REGION"]
        )
        s3client = boto3.client("s3", **credentials)
        return s3client, s3region



# pylint: disable=W0201, E1101
""" handle request for markdown pages """
import logging
import os
import importlib
from tornado.web import RequestHandler, HTTPError
from tornado.escape import url_escape
from ..utils.converter_mixin import ConverterMixin
from .access_control import UserMixin
from ..utils.nav import nav


LOGGER = logging.getLogger(__name__)
EMPTY_TOC = '<div class="toc">\n<ul></ul>\n</div>\n'


class SiteHandler(
    UserMixin, ConverterMixin, RequestHandler
):  # pylint: disable=W0223
    """ inline transform request for markdown pages """

    def initialize(self, pages):
        """ setup init properties """
        self.pages = pages
        self.meta = None
        self.nav = None
        self.site_nav = None
        self.site = None

    def create_template_loader(self, template_path):
        """ if we have one, us it """
        if self.site.template_loader:
            return self.site.template_loader
        return super().create_template_loader(template_path)

    @property
    def has_toc(self):
        """ determin if toc is empty """
        return self.meta.toc != EMPTY_TOC

    def meta_value(self, name, default=None):
        """ return markdown meta value """
        return self.meta.Meta.get(name, [default])

    def one_meta_value(self, name, default=None):
        """ return markdown meta value """
        result = self.meta_value(name, default)
        return result[0] if result else None

    def load_site_nav(self, site, path):
        """ set the handler site_nav attribute """
        menu = nav(site, root=self.pages, path=path)
        if menu:
            self.site_nav = "\n".join(menu)

    def load_dir_nav(self, site, path):
        """ load nav section if it exist """
        folder = os.path.dirname(path)
        if folder:
            LOGGER.info(" -- folder: %s", folder)
            nav_path = os.path.join(folder, "-nav.md")
            _, content = site.get_file(nav_path)
            if content:
                content = content.decode("utf-8")
                LOGGER.info(" -- nav: %s", nav_path)
                content = self.meta.convert(content)
                self.nav = self.convert_images(content)

    def run_script(
        self, site, script_name, path
    ):  # pylint: disable=unused-argument
        """ load a module and call module.main """
        name = f"{self.settings['script_path']}.{script_name}"
        script_module = importlib.import_module(name)

        return script_module.main(path)

    async def get(self, path):
        """ handle get """
        path = path if path else "index.html"

        file, ext = os.path.splitext(path)

        doc = os.path.join(self.pages, f"{file}.md")
        self.site = self.get_site(path)
        _, content = self.site.get_file(doc)
        if content is None:
            raise HTTPError(404)
        if content:
            content = content.decode("utf-8")

        self.meta = self.markdown
        self.load_dir_nav(self.site, doc)
        self.load_site_nav(self.site, path)

        file_path = os.path.split(file)[0]

        # load theme
        theme_file = os.path.join(self.pages, file_path, "-theme.css")
        _, theme_css = self.site.get_file(theme_file)
        if theme_css:
            LOGGER.info(" -- theme.css")
            theme_css = theme_css.decode("utf-8")

        edit_path = "/edit"
        if file:
            edit_path = f"/edit?path={ url_escape(file) }.md"

        LOGGER.info(" -- ext: %s", ext)
        if ext == ".html":
            content = self.meta.convert(content)
            LOGGER.info(" -- meta: %s", self.meta.Meta)
            template = self.one_meta_value("template", "site")
            LOGGER.info(" -- tmpl: %s", template)
            for key in self.meta.Meta:
                if key.startswith("x-script-"):
                    outcome = self.run_script(
                        self.site, self.meta.Meta[key][0], path
                    )
                    self.meta.Meta[key] = [outcome]
            self.render(
                f"{template}_tmpl.html",
                content=self.convert_images(content),
                edit_path=edit_path,
                theme_css=theme_css,
            )
        else:
            self.write(self.convert_images(content))



# pylint: disable=attribute-defined-outside-init
""" Proxy S3 as static file handler """
import logging
from tornado.web import StaticFileHandler
from .access_control import UserMixin

LOGGER = logging.getLogger(__name__)


class StaticFiles(UserMixin, StaticFileHandler):  # pylint: disable=W0223
    """ work around """

    sites = {}

    @classmethod
    def get_content(cls, abspath, start=None, end=None):
        """ return content """
        LOGGER.info("get abs: %r", abspath)
        site = cls.sites[abspath]
        _, data = site.get_file(abspath)
        return data

    def _stat(self):
        assert self.absolute_path is not None
        abspath = self.absolute_path
        LOGGER.info("static abs: %r %r", self.root, self.absolute_path)
        if not hasattr(self, "_stat_result"):
            result = self.site.get_head(abspath)
            self._stat_result = result  # pylint: disable=W0201
        return self._stat_result

    @classmethod
    def get_absolute_path(cls, root, path):
        """ return abs path of content """
        LOGGER.info("make abs: %r %r", root, path)
        root = root[:-1] if root[-1] == "/" else root
        return path

    def validate_absolute_path(self, root, absolute_path):
        """ is it valid? and cache site for abspath """
        LOGGER.info("validate abs: %r %r", root, absolute_path)
        static_prefix = self.settings["static_prefix"]
        current_user = self.get_current_user()
        self.site = self.application.get_site(
            current_user, absolute_path, self.request
        )
        abspath = f"{static_prefix}{absolute_path}"
        self.sites[abspath] = self.site
        LOGGER.info("site set: %r", abspath)
        return abspath

    def finish(self, chunk=None):
        """ tidy up sites """
        result = super().finish(chunk)
        self.site = None
        if self.absolute_path:
            del self.sites[self.absolute_path]
        return result



# pylint: disable=W0201, W0223
""" browser and upload to bucket """
import os
import logging
import tornado.web
from .base_handler import BaseHandler
from .access_control import UserMixin
from ..utils.json_utils import dumps

LOGGER = logging.getLogger(__name__)

TYPE_MAP = {
    ".svg": ("SVG", "image/svg+xml"),
    ".jpg": ("JPEG", "image/jpeg"),
    ".jpeg": ("JPEG", "image/jpeg"),
    ".png": ("PNG", "image/png"),
    ".gif": ("GIF", "image/gif"),
}


class S3Browser(UserMixin, BaseHandler):
    """ list contents of bucket """

    def initialize(
        self,
        bucket_name=None,
        folder="",
    ):
        """ setup s3 bucket """
        self.s3bucket = bucket_name
        self.folder = folder

    @property
    def img_path(self):
        """ return application img_path """
        return self.settings.get("img_path")

    @property
    def local_images(self):
        """ return application local_images """
        return self.settings.get("local_images")

    def add(self, body, path, meta=None):
        """ adds body, returns path """
        key = self.folder + path
        LOGGER.info("adding %s", key)
        site = self.get_site(key)
        return site.put_file(body=body, key=key, meta=meta)

    @tornado.web.authenticated
    def get(self, prefix=None):
        """ returns the contents of bucket """
        result = None
        if prefix is None:
            prefix = ""
        LOGGER.info("browse: %s", prefix)
        self.set_header("Content-Type", "application/json")
        prefix = self.folder + prefix if self.folder else prefix
        LOGGER.info("prefix: %s", prefix)
        site = self.get_site(prefix)
        result = site.list_folder(prefix, self.folder)
        self.write(dumps(result))

    @tornado.web.authenticated
    def put(self, path=None):  # pylint: disable=W0613
        """ return the img_path """
        self.write({"img_path": self.img_path})

    @tornado.web.authenticated
    def post(self, path=None):
        """ handle the upload """
        result = []
        for key in self.request.files:
            for fileinfo in self.request.files[key]:
                fname = fileinfo["filename"]
                _, ext = os.path.splitext(fname)
                ftype, fmime = TYPE_MAP.get(ext.lower(), (None, None))
                if ftype is None:
                    raise Exception("File Type not accepted: {}".format(ftype))
                LOGGER.info("upload: %s", f"{path}{fname}")
                s3key = self.add(
                    fileinfo["body"],
                    path=f"{path}{fname}",
                    meta={"original_name": fname, "content-type": fmime},
                )
                result.append(s3key)
        self.write({"result": result})



# pylint: disable=too-many-instance-attributes, too-few-public-methods
""" Duckdown configuration """
import os
import time
import logging
from pkg_resources import resource_filename

LOGGER = logging.getLogger(__name__)


class Config:
    """ holding all the variables """

    # constants
    PAGE_PATH = "pages/"
    TEMPLATE_PATH = "templates/"
    IMAGES_PATH = "static/images/"
    SCRIPT_PATH = "scripts/"
    STATIC_PATH = "static/"
    USERS_PATH = "users.json"
    IMG_PATH = "/static/images/"
    ASSETS_PREFIX = "/edit/assets/"

    # variables
    debug = False
    port = int(os.getenv("PORT", "8080"))
    bucket_name = ""
    credentials = {}
    app_path = ""
    app_name = "duckdown_app"

    static_prefix = STATIC_PATH
    image_path = IMAGES_PATH
    image_bucket = None
    users_path = USERS_PATH

    image_credentials = {}
    img_path = IMG_PATH
    convert_image_paths = True

    duck_assets_prefix = ASSETS_PREFIX
    duck_assets = resource_filename("duckdown", "assets")
    duck_templates = resource_filename("duckdown", "templates")

    cookie_secret = "it was a dark and stormy duckdown"
    login_url = "/login"
    login_handler = None

    vue_page = "vue.html"
    vue_src = "./client/src/"
    vue_assets = resource_filename("duckdown", "assets/vue/")
    vue_manifest = resource_filename("duckdown", "assets/vue/manifest.json")

    def __init__(self, app_path=None, bucket=None, debug=None, port=None):
        """ init derived data members """

        # allow for command line override
        if app_path is not None:
            self.app_path = app_path
        if bucket is not None:
            self.bucket_name = bucket
        if debug is not None:
            self.debug = debug
        if port is not None:
            self.port = port

        LOGGER.debug("init app_path: %s", self.app_path)

        self.static_path = os.path.join(self.app_path, self.STATIC_PATH)
        self.template_path = os.path.join(self.app_path, self.TEMPLATE_PATH)
        self.script_path = os.path.join(self.app_path, self.SCRIPT_PATH)
        self.page_path = os.path.join(self.app_path, self.PAGE_PATH)
        self.cookie_name = f"{self.app_name}-user"

    def tornado_settings(self, settings):
        """ setup tornado settings dict """

        settings.setdefault("debug", self.debug)
        settings.setdefault("port", self.port)
        settings.setdefault("app_name", self.app_name)
        settings.setdefault("app_path", self.app_path)

        # site paths
        settings.setdefault("users_path", self.users_path)
        settings.setdefault("static_path", self.static_path)
        settings.setdefault("static_prefix", self.static_prefix)
        settings.setdefault("template_path", self.template_path)
        settings.setdefault("script_path", self.script_path)
        settings.setdefault("page_path", self.page_path)
        settings.setdefault("img_path", self.img_path)
        settings.setdefault("convert_image_paths", self.convert_image_paths)

        # editor setup
        settings.setdefault("duck_assets_prefix", self.duck_assets_prefix)
        settings.setdefault("duck_assets", self.duck_assets)
        settings.setdefault("duck_templates", self.duck_templates)

        # access control
        settings.setdefault("cookie_name", f"{self.app_name}-user")
        if settings.get("debug") is True:
            settings[
                "cookie_secret"
            ] = f"it was a dark and stormy duckdown {time.time()}"
        else:
            settings.setdefault("cookie_secret", self.cookie_secret)
        settings.setdefault("login_url", self.login_url)

        return settings



# pylint: disable=W0613
""" we want to create a commonality with directory """
import os
import logging
import mimetypes
from .head import Head
from .tmpl_loader import TmplLoader

LOGGER = logging.getLogger(__name__)


class Folder:
    """ files and folders """

    def __init__(self, directory, subdirectory=""):
        self.directory = directory
        self.subdirectory = subdirectory
        self.template_loader = TmplLoader(self)
        self.image_bucket = self

    def for_subfolder(self, value):
        """ return a clone for a subfolder """
        return Folder(self.directory, subdirectory=value)

    def make_path(self, path):
        """ return path """
        return os.path.join(self.directory, self.subdirectory, path)

    def set_image_bucket(self, value):
        """ set the image bucket """
        self.image_bucket = value

    def is_file(self, path):
        """ is this a file """
        path = self.make_path(path)
        return os.path.isfile(path)

    def get_head(self, key):
        """ return Head on key """
        path = os.path.join(self.directory, self.subdirectory, key)
        stat = os.stat(path)
        return Head(path=path, st_size=stat.st_size, st_mtime=stat.st_mtime)

    def get_file(self, key):
        """ returns file key in directory """
        result = (None, None)
        path = os.path.join(self.directory, self.subdirectory, key)
        if os.path.isfile(path):
            content_type, _ = mimetypes.guess_type(path)
            with open(path, "rb") as file:
                result = content_type, file.read()
        return result

    def put_file(self, body, key, **kwargs):
        """ put file into directory """
        path = os.path.join(self.directory, self.subdirectory, key)
        folder, _ = os.path.split(path)
        if folder and not os.path.exists(folder):
            LOGGER.info("making directory: %s", folder)
            os.makedirs(folder)
        with open(path, "wb") as file:
            file.write(body)
        return path

    def delete_file(self, key):
        """ will remove file from directory """
        path = os.path.join(self.directory, self.subdirectory, key)
        if os.path.isfile(path):
            os.unlink(path)
        else:
            raise ValueError(f"No such file: {key}")

    def list_folder(self, prefix="", root="", delimiter="/"):
        """ list the contents of folder """
        LOGGER.info("listing: (%s, %s)", root, prefix)
        abspath = self.make_path(prefix)
        LOGGER.info("listing abs: %s", abspath)
        absroot = self.make_path(root)
        starts = len(absroot)
        files = []
        folders = []
        if os.path.isdir(abspath):
            with os.scandir(abspath) as item:
                for entry in item:
                    is_file = entry.is_file()
                    size = entry.stat().st_size if is_file else None
                    mime = (
                        mimetypes.guess_type(entry.path) if is_file else None
                    )
                    if entry.name[0] != ".":
                        item = {
                            "name": entry.name,
                            "path": entry.path[starts:],
                            "file": is_file,
                            "size": size,
                            "type": mime,
                        }
                        if is_file:
                            files.append(item)
                        else:
                            folders.append(item)
        return {"files": files, "folders": folders}



# pylint: disable=W0201
""" Handle request for index page """
from tornado.web import RequestHandler, authenticated
from .access_control import UserMixin
from ..utils.assets_mixin import AssetsMixin


class EditorHandler(
    AssetsMixin, UserMixin, RequestHandler
):  # pylint: disable=W0223
    """ return index page """

    def initialize(self, manifest, page):
        """ setup production """
        self.manifest = manifest
        self.page = page

    @authenticated
    def get(self):
        """ handle get request """
        self.render(self.page)



""" we want to create a commonality with directory """
import io
import logging
import mimetypes
import time
import botocore
from .head import Head
from .tmpl_loader import TmplLoader


LOGGER = logging.getLogger(__name__)


class S3Folder:
    """ files and folders """

    def __init__(self, bucket_name, client, region, subfolder=""):
        self.s3region = region
        self.s3client = client
        self.s3bucket = bucket_name
        self.subfolder = subfolder
        self.s3bucket_url = (
            f"//s3-{self.s3region}.amazonaws.com/{bucket_name}{subfolder}"
        )
        self.template_loader = TmplLoader(self)
        self.image_bucket = self

    def for_subfolder(self, value):
        """ return a clone for a subfolder """
        return S3Folder(
            self.s3bucket,
            client=self.s3client,
            region=self.s3region,
            subfolder=value,
        )

    def set_image_bucket(self, value):
        """ set the image bucket """
        self.image_bucket = value

    @classmethod
    def _last_item_(cls, value, sep="/"):
        if sep in value:
            if value[-1] == sep:
                value = value[:-1].split(sep)[-1]
            value = value.split(sep)[-1]
        return value

    @classmethod
    def is_file(cls, path):
        """ is this a file """
        return path[-1] != "/"

    @classmethod
    def scan_path(cls, client, bucket, prefix, delimiter="/"):
        """ return custom format of listing """

    def list_folder(self, prefix="", root="", delimiter="/"):
        """ list the contents of bucket folder """
        response = self.s3client.list_objects_v2(
            Bucket=self.s3bucket, Prefix=prefix, Delimiter=delimiter
        )
        starts = len(root)
        LOGGER.debug(response)
        result = {
            "files": [
                {
                    "path": item["Key"][starts:],
                    "name": self._last_item_(item["Key"]),
                    "size": item["Size"],
                    "type": mimetypes.guess_type(item["Key"])[0],
                    "file": True,
                }
                for item in response.get("Contents", [])
            ],
            "folders": [
                {
                    "path": item["Prefix"][starts:],
                    "name": self._last_item_(item["Prefix"]),
                    "file": False,
                }
                for item in response.get("CommonPrefixes", [])
            ],
        }
        return result

    def get_head(self, key):
        """ return meta data on key """
        response = self.s3client.head_object(Bucket=self.s3bucket, Key=key)
        LOGGER.debug(response)
        st_size = int(response["ContentLength"])
        dt_obj = response["LastModified"]
        st_mtime = time.mktime(dt_obj.timetuple())
        return Head(path=key, st_size=st_size, st_mtime=st_mtime)

    def get_file(self, key):
        """ returns object of key """
        content_type, content = None, None
        try:
            LOGGER.info("getting: %s", key)
            data = io.BytesIO()
            self.s3client.download_fileobj(
                Bucket=self.s3bucket, Key=key, Fileobj=data
            )
            content_type, _ = mimetypes.guess_type(key)
            content = data.getvalue()
        except botocore.exceptions.ClientError as err:
            if err.response["ResponseMetadata"]["HTTPStatusCode"] != 404:
                LOGGER.info(err)
                raise
        return content_type, content

    def put_file(
        self, body=None, key=None, meta=None, mime=None, enc=None
    ):  # pylint: disable=R0913
        """ put_object """
        if mime:
            content_type, content_encoding = mime, enc
        else:
            content_type, content_encoding = mimetypes.guess_type(key)
        if content_type is None:
            raise Exception(f"Unknown Content Type: {key}")

        kwargs = {
            "ACL": "public-read",
            "Body": body,
            "Bucket": self.s3bucket,
            "Key": key,
            "ContentType": content_type,
        }
        if meta:
            kwargs["Metadata"] = meta
        if content_encoding:
            kwargs["ContentEncoding"] = content_encoding

        response = self.s3client.put_object(**kwargs)
        LOGGER.debug(response)
        result = self.s3client.generate_presigned_url(
            "get_object", Params={"Bucket": self.s3bucket, "Key": key}
        )
        return result

    def delete_file(self, key):
        """ will remove file from bucket """
        response = self.s3client.delete_object(Bucket=self.s3bucket, Key=key)
        LOGGER.debug(response)
        return response



""" an example function and exposing it through __all__ """
import base64
import os
from contextlib import contextmanager
from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC


@contextmanager
def fkey(envar="DKDN_KEY", passcode=None):
    """ will provide crypto using os.environ[envar] """

    try:
        key = os.environ[envar]
    except KeyError as ex:
        raise KeyError(f"Expecting environment variable {envar}") from ex

    if passcode:
        password = passcode.encode()  # Convert to type bytes
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=32,
            salt=key.encode(),
            iterations=100000,
            backend=default_backend(),
        )
        key = base64.urlsafe_b64encode(
            kdf.derive(password)
        )  # Can only use kdf once
    crypto = Fernet(key)
    yield crypto


def encrypt(message, passcode=None):
    """ returns an encrytped utf-8 string """
    with fkey(passcode=passcode) as crypto:
        return crypto.encrypt(message.encode()).decode("utf-8")


def decrypt(message, passcode=None):
    """ returns a decrypted utf-8 string """
    with fkey(passcode=passcode) as crypto:
        return crypto.decrypt(message.encode()).decode("utf-8")



# pylint: disable=W0201, W0223
""" We want to manage a directory of files """
import os
import logging
import mimetypes
from .base_handler import BaseHandler
from .access_control import UserMixin

LOGGER = logging.getLogger(__name__)
mimetypes.add_type("text/markdown", ".md")


class DirHandler(UserMixin, BaseHandler):
    """ Manage a directory """

    def initialize(self, directory):
        """ setup directory """
        self.directory = directory

    def get(self, path=None):
        """ return the files and directories in path """
        path = os.path.join(self.directory, path) if path else self.directory
        site = self.get_site(path)
        LOGGER.info("dir get path: %s", path)
        if site.is_file(path):
            LOGGER.info("loading file: %s", path)
            content_type, body = site.get_file(path)
            if content_type:
                self.set_header("Content-Type", content_type)
            self.write(body)
        else:
            LOGGER.info("listing folder: %s", path)
            self.write(site.list_folder(path))

    def put(self, path):
        """ handle the setting of file to path """
        LOGGER.info("saving %s", path)
        mime, _ = mimetypes.guess_type(path)
        path = os.path.join(self.directory, path)
        site = self.get_site(path)
        site.put_file(self.request.body, path, mime=mime)
        self.write("saved")

    def delete(self, path):
        """ will remove a document """
        path = os.path.join(self.directory, path)
        site = self.get_site(path)
        site.delete_file(path)
        self.write("deleted")



""" application entry point """
import os
import logging
import tornado.ioloop
import tornado.log

LOGGER = logging.getLogger(__name__)


def run(app):
    """ make an app and run it """
    if app.settings["debug"] is True:
        LOGGER.info("running in debug mode")

    port = os.getenv("PORT", app.settings.get("port", "8080"))
    app.listen(port)
    LOGGER.info("listening on port: %s", port)

    ioloop = tornado.ioloop.IOLoop.current()
    try:
        ioloop.start()
    except KeyboardInterrupt:
        LOGGER.info("shutting down")
        ioloop.stop()



""" task to create app template """
import os
import shutil
import logging
from shutil import copyfile
from pkg_resources import resource_filename
from invoke import task
from .provision import s3_tools

LOGGER = logging.getLogger(__name__)
SOURCES = [
    ("", "users.json"),
    ("pages", "index.md"),
    ("static", "site.css"),
    ("static", "favicon.ico"),
    ("static", "robots.txt"),
    ("static/images", "logo.svg"),
    ("templates", "site_tmpl.html"),
]


def populate_folder(path, force=False):
    """ create a duckdown app at path """
    if os.path.exists(path):
        if force is False:
            LOGGER.info("already exists: %s", path)
            return

        LOGGER.info("removing %s", path)
        shutil.rmtree(path)

    LOGGER.info("creating %s", path)
    for folder in {folder for folder, _ in SOURCES if folder}:
        os.makedirs(os.path.join(path, folder), exist_ok=True)

    for folder, file in SOURCES:
        copyfile(
            resource_filename("duckdown.tool", f"data/{file}"),
            os.path.join(path, folder, file),
        )


def populate_bucket(bucket, force=False):
    """ create a duckdown app in bucket """

    if force is True:
        LOGGER.info("emptying: %s", bucket)
        s3_tools.empty_bucket(bucket)

    LOGGER.info("loading: %s", bucket)
    for folder, file in SOURCES:
        if file == "config.ini":
            file = "s3_config.ini"
        key = f"{folder}/{file}" if folder else file
        path = resource_filename("duckdown.tool", f"data/{file}")
        s3_tools.upload(bucket, key, path)

    LOGGER.info("done!")


@task
def create(_, path, force=False):
    """ create a duckdown app at path """
    return populate_folder(path, force)


@task
def s3_create(_, bucket, force=False):
    """ create a duckdown app in bucket """
    return populate_bucket(bucket, force)



""" an s3 template loader """
import os
import logging
from typing import Any
from tornado.template import BaseLoader, Template

LOGGER = logging.getLogger(__name__)


class TmplLoader(BaseLoader):
    """A template loader that loads from a single root directory."""

    def __init__(self, site, folder="templates/", **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.site = site
        self.folder = folder

    def resolve_path(self, name, parent_path=None):
        """ resolve a path """
        if (
            parent_path
            and not parent_path.startswith("<")
            and not parent_path.startswith("/")
            and not name.startswith("/")
        ):
            current_path = os.path.join(self.folder, parent_path)
            file_dir = os.path.dirname(os.path.abspath(current_path))
            relative_path = os.path.abspath(os.path.join(file_dir, name))
            if relative_path.startswith(self.folder):
                name = relative_path[len(self.folder) + 1 :]
        return name

    def _create_template(self, name: str) -> Template:
        """ create a template from bucket/folder/name """
        key = f"{self.folder}{name}"
        LOGGER.info("template: %s", key)
        _, data = self.site.get_file(key)
        template = Template(data.decode("utf-8"), name=name, loader=self)
        return template



"""
    This is a our invoke program
"""
import os
import sys
import tornado.log
from invoke import Program, Collection
from duckdown import VERSION
from .create import create, s3_create
from .run import run
from .publish import publish
from .secure import secure, unsecure, gen_key

sys.path.insert(0, os.getcwd())

VERSION = "0.0.12"

_NAMESPACE_ = Collection()

_NAMESPACE_.add_task(create)
_NAMESPACE_.add_task(run)
_NAMESPACE_.add_task(publish)
_NAMESPACE_.add_task(secure)
_NAMESPACE_.add_task(unsecure)
_NAMESPACE_.add_task(gen_key)
_NAMESPACE_.add_task(s3_create)

tornado.log.enable_pretty_logging()

program = Program(
    version=VERSION, namespace=_NAMESPACE_
)  # pylint: disable=C0103



# pylint: disable=R0913
""" run duckdown app """
import logging
from invoke import task
from duckdown.app import App
from duckdown.config import Config
from duckdown.utils import run_tornado

LOGGER = logging.getLogger(__name__)


@task
def run(
    _,
    app_path="",
    bucket="",
    debug=False,
    port=8080,
):
    """ run app """
    settings = {
        "app_path": app_path,
        "bucket": bucket,
        "debug": debug,
        "port": port,
    }
    if not app_path and not bucket:
        LOGGER.info("run: %s", settings)
        print("either a path or bucket are required!")
    else:
        app = App(Config(**settings))
        run_tornado.run(app)
